-- VehicleVaulting. Factorio mod: Enter vehicles anywhere, at any time from your inventory without placing them first. Exiting picks it up again automatically.
-- Copyright (C) 2016  Qon

-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>

local debug = true
local no_vaulting_start = {['diesel-locomotive'] = true}

function init()
    if global.players_data == nil then
        global.players_data = {}
        for i=1, #game.players do
            init_player(i, false)
        end
    end
    global.vehicles = global.vehicles or {}
end

function init_player(player_index, force)
    init_player_settings(player_index, force)
    -- global.players_data[player_index].autostarted = nil
end

function init_player_settings(player_index, force)
    init(nil) -- TODO Shouldn't be necessary if the migration was done correctly...
    if global.players_data[player_index] == nil or force then
        global.players_data[player_index] = {}
    end
    if global.players_data[player_index].prev_pos == nil then
        global.players_data[player_index].prev_pos = game.players[player_index].position
    end
end

function vault(event)
    local p = game.players[event.player_index]
    init_player(event.player_index, false)
    local d = global.players_data[event.player_index]
    if p.vehicle == nil then
        d.vault_in_tick = game.tick
    else
        d.vault_out_tick = game.tick
    end
    global.players_data[event.player_index] = d
    -- if debug then p.print('vault event '..game.tick) end
end

function driving_changed_state(event)
    local p = game.players[event.player_index]
    -- if debug then p.print('driving_changed_state event '..game.tick) end
    init()
    d = global.players_data[event.player_index]
    if p.vehicle == nil then 
        if d.v.valid and d.last_vehicle_pos ~= nil and d.vd.name ~= 'diesel-locomotive' and d.vault_out_tick == game.tick then
            d.v.destroy()
            p.teleport(d.last_vehicle_pos)
            d.v = nil
        end
    else
        d.v = p.vehicle
        d.vd = {name = d.v.name}
        -- if debug then p.print('Added '..p.vehicle.name..' to the list of vaultable vehicles.  ') end
        global.vehicles = global.vehicles or {}
        global.vehicles[p.vehicle.name] = true
    end
    global.players_data[event.player_index] = d
end

    -- function exp(b, e)
    --     return math.exp(math.log(b)*e)
    -- end

function distance(a, b)
    local A = a.x - b.x
    A = A*A
    local B = a.y - b.y
    B = B*B
    return math.sqrt(A + B)
end

function tick(event)
    for i=1, #game.players do
        local p = game.players[i]
        -- if debug then p.print('tick event '..game.tick) end
        init_player(i)
        local d = global.players_data[i]
        if p.vehicle == nil and p.walking_state.walking then --and (p.position.x ~= d.prev_pos.x or p.position.y ~= d.prev_pos.y) then
            d.last_walk_direction = p.walking_state.direction
            d.deltapos = {x = p.position.x - d.prev_pos.x, y = p.position.y - d.prev_pos.y}
            d.moving = game.tick
            -- if debug then p.print('deltapos('..d.deltapos.x..', '..d.deltapos.y..')') end
        end
        if p.vehicle ~= nil then
            d.last_vehicle_pos = p.position
        end
        if p.connected then
            -- if debug then 
            --     local dir = game.tick%360
            --     p.print('test direction('..dir..', '..math.deg(math.atan2(math.sin(math.rad(dir)), math.cos(math.rad(dir))))..')')
            -- end

            -- Vault into vehicle if we didn't just enter a vehicle
            if d.vault_in_tick ~= nil and d.vault_in_tick + 1 == game.tick and p.vehicle == nil then
                local holding_vehicle
                if p.cursor_stack.valid_for_read and global.vehicles[p.cursor_stack.name] then
                    d.vd = {name = p.cursor_stack.name}
                end
                if d.vd ~= nil or holding_vehicle then
                    local direction = defines.direction.north
                    -- if debug then p.print('deltapos('..d.deltapos.x..', '..d.deltapos.y..')') end
                    -- if debug then p.print('direction('..direction..', '..math.atan(-d.deltapos.y, d.deltapos.x)..')') end
                    -- if debug then p.print('pos('..p.position.x..', '..p.position.y..')') end
                    local vehicle = p.surface.create_entity{name = d.vd.name, position = p.position, force = p.force, direction = d.last_walk_direction}
                    if vehicle ~= nil then    
                        vehicle.passenger = p.character
                        vehicle.insert{name = 'coal', count = 50}
                        if d.moving ~= nil and d.moving > game.tick - 10 and not no_vaulting_start[d.vd.name] then
                            vehicle.speed = distance({x = 0, y = 0}, d.deltapos) -- game.entity_prototypes[vehicle.name].max_speed
                        end
                        game.raise_event(defines.events.on_player_driving_changed_state, {player_index = i})
                    end
                    p.remove_item{name = d.vd.name, count = 1}
                else
                    if d.vd == nil then
                        p.print('Can\'t vault into vehicle. You have never entered a vehicle before.')
                    else
                        p.print('You can\'t vault right now. Reason unknown.')
                    end
                end
            -- Vault out of the vehicle.
            elseif d.vault_out_tick ~= nil and d.vault_out_tick + 1 == game.tick then
                local vehicle = p.vehicle or d.v
                if vehicle ~= nil and vehicle.valid and (vehicle.name ~= 'diesel-locomotive' or (#vehicle.train.cargo_wagons == 0 and #vehicle.train.locomotives.front_movers == 1 and #vehicle.train.locomotives.back_movers == 0)) then
                    p.insert{name = vehicle.name, count = 1, health = vehicle.health/game.entity_prototypes[vehicle.name].max_health}
                    local pos = p.position
                    vehicle.passenger = nil
                    p.teleport(pos)
                    vehicle.destroy()
                    game.raise_event(defines.events.on_player_driving_changed_state, {player_index = i})
                end
            end
        end
        d.prev_pos = {x = p.position.x, y = p.position.y}
        global.players_data[i] = d
    end
end
script.on_event(defines.events.on_tick, tick)
script.on_event(defines.events.on_player_joined_game, function(event) init_player(event.player_index, false) end)
script.on_event(defines.events.on_player_driving_changed_state, driving_changed_state)
script.on_event('vehiclevaulting-vault', vault)